import logging
import os

def setup_logger(name: str) -> logging.Logger:
    logger = logging.getLogger(name)
    if os.environ.get("MULTICALL_DEBUG", False):
        logger.addHandler(logging.StreamHandler())
        logger.setLevel(logging.DEBUG)
    return logger




import asyncio
from concurrent.futures import ProcessPoolExecutor
from typing import Any, Awaitable, Callable, Coroutine, Dict, Iterable

import eth_retry
from aiohttp import ClientTimeout
from web3 import AsyncHTTPProvider, Web3
from web3.eth import AsyncEth
from web3.providers.async_base import AsyncBaseProvider

from multicall.constants import AIOHTTP_TIMEOUT, NUM_PROCESSES, NO_STATE_OVERRIDE

chainids: Dict[Web3,int] = {}

@eth_retry.auto_retry
def chain_id(w3: Web3) -> int:
    '''
    Returns chain id for an instance of Web3. Helps save repeat calls to node.
    '''
    try:
        return chainids[w3]
    except KeyError:
        chainids[w3] = w3.eth.chain_id
        return chainids[w3]

async_w3s: Dict[Web3,Web3] = {}
process_pool_executor = ProcessPoolExecutor(NUM_PROCESSES)

def get_endpoint(w3: Web3) -> str:
    provider = w3.provider
    if isinstance(provider, str):
        return provider
    if hasattr(provider, "_active_provider"):
        provider = provider._get_active_provider(False)
    return provider.endpoint_uri

def get_async_w3(w3: Web3) -> Web3:
    if w3 in async_w3s:
        return async_w3s[w3]
    if w3.eth.is_async and isinstance(w3.provider, AsyncBaseProvider):
        timeout = w3.provider._request_kwargs["timeout"]
        if isinstance(timeout, ClientTimeout):
            timeout = timeout.total

        if timeout < AIOHTTP_TIMEOUT.total:
            w3.provider._request_kwargs["timeout"] = AIOHTTP_TIMEOUT

        async_w3s[w3] = w3
        return w3
    request_kwargs = {'timeout': AIOHTTP_TIMEOUT}
    async_w3 = Web3(
        provider=AsyncHTTPProvider(get_endpoint(w3), request_kwargs),
        # In older web3 versions, AsyncHTTPProvider objects come
        # with incompatible synchronous middlewares by default.
        middlewares=[],
    )
    async_w3.eth = AsyncEth(async_w3)
    async_w3s[w3] = async_w3
    return async_w3

def get_event_loop() -> asyncio.AbstractEventLoop:
    try:
        return asyncio.get_event_loop()
    except RuntimeError as e: # Necessary for use with multi-threaded applications.
        if not str(e).startswith("There is no current event loop in thread"):
            raise e
        return asyncio.new_event_loop()

def await_awaitable(awaitable: Awaitable) -> Any:
    return get_event_loop().run_until_complete(awaitable)

async def run_in_subprocess(callable: Callable, *args: Any, **kwargs) -> Any:
    if NUM_PROCESSES == 1:
        return callable(*args, **kwargs)
    return await asyncio.get_event_loop().run_in_executor(process_pool_executor, callable, *args, **kwargs)

def raise_if_exception(obj: Any) -> None:
    if isinstance(obj, Exception):
        raise obj

def raise_if_exception_in(iterable: Iterable[Any]) -> None:
    for obj in iterable:
        raise_if_exception(obj)

async def gather(coroutines: Iterable[Coroutine]) -> None:
    results = await asyncio.gather(*coroutines, return_exceptions=True)
    raise_if_exception_in(results)
    return results

def state_override_supported(w3: Web3) -> bool:
    if chain_id(w3) in NO_STATE_OVERRIDE:
        return False
    return True



import asyncio
from time import time
from typing import Any, Dict, List, Optional, Tuple, Union

import aiohttp
import requests
from web3 import Web3

from multicall import Call
from multicall.constants import (ASYNC_SEMAPHORE, GAS_LIMIT,
                                 MULTICALL2_ADDRESSES, MULTICALL3_ADDRESSES,
                                 MULTICALL3_BYTECODE, w3)
from multicall.loggers import setup_logger
from multicall.utils import (await_awaitable, chain_id, gather,
                             run_in_subprocess, state_override_supported)

logger = setup_logger(__name__)

CallResponse = Tuple[Union[None,bool],bytes]

def get_args(calls: List[Call], require_success: bool = True) -> List[Union[bool,List[List[Any]]]]:
    if require_success is True:
        return [[[call.target, call.data] for call in calls]]
    return [require_success, [[call.target, call.data] for call in calls]]

def unpack_aggregate_outputs(outputs: Any) -> Tuple[CallResponse,...]:
    return tuple((None, output) for output in outputs)

def unpack_batch_results(batch_results: List[List[CallResponse]]) -> List[CallResponse]:
    return [result for batch in batch_results for result in batch]


class Multicall:
    def __init__(
        self, 
        calls: List[Call],
        block_id: Optional[int] = None, 
        require_success: bool = True,
        gas_limit: int = GAS_LIMIT,
        _w3: Web3 = w3
    ) -> None:
        self.calls = calls
        self.block_id = block_id
        self.require_success = require_success
        self.gas_limit = gas_limit
        self.w3 = _w3
        self.chainid = chain_id(self.w3)
        if require_success is True:
            multicall_map = MULTICALL3_ADDRESSES if self.chainid in MULTICALL3_ADDRESSES else MULTICALL2_ADDRESSES
            self.multicall_sig = 'aggregate((address,bytes)[])(uint256,bytes[])'
        else:
            multicall_map = MULTICALL3_ADDRESSES if self.chainid in MULTICALL3_ADDRESSES else MULTICALL2_ADDRESSES
            self.multicall_sig = 'tryBlockAndAggregate(bool,(address,bytes)[])(uint256,uint256,(bool,bytes)[])'
        self.multicall_address = multicall_map[self.chainid]

    def __call__(self) -> Dict[str,Any]:
        start = time()
        response = await_awaitable(self)
        logger.debug(f"Multicall took {time() - start}s")
        return response
     
    def __await__(self) -> Dict[str,Any]:
        return self.coroutine().__await__()

    async def coroutine(self) -> Dict[str,Any]:
        batches = await gather([
            self.fetch_outputs(batch, id=str(i)) 
            for i,batch in enumerate(batcher.batch_calls(self.calls,batcher.step))
        ])
        outputs = await run_in_subprocess(unpack_batch_results, batches)

        return {
            name: result
            for output in outputs
            for name, result in output.items()
        }

    async def fetch_outputs(self, calls: List[Call], ConnErr_retries: int = 0, id: str = '') -> List[CallResponse]:
        logger.debug(f"coroutine {id} started")

        if calls is None:
            calls = self.calls
        
        async with ASYNC_SEMAPHORE:
            try:
                args = await run_in_subprocess(get_args, calls, self.require_success)
                if self.require_success is True:
                    _, outputs = await self.aggregate.coroutine(args)
                    outputs = await run_in_subprocess(unpack_aggregate_outputs, outputs)
                else:
                    _, _, outputs = await self.aggregate.coroutine(args)
                outputs = await gather([
                    run_in_subprocess(Call.decode_output, output, call.signature, call.returns, success)
                    for call, (success, output) in zip(calls, outputs)
                ])
                logger.debug(f"coroutine {id} finished")
                return outputs
            except Exception as e:
                _raise_or_proceed(e, len(calls), ConnErr_retries=ConnErr_retries)
        
        # Failed, we need to rebatch the calls and try again.
        batch_results = await gather([
            self.fetch_outputs(chunk, ConnErr_retries+1, f"{id}_{i}")
            for i, chunk in enumerate(await batcher.rebatch(calls))
        ])
            
        return_val = await run_in_subprocess(unpack_batch_results,batch_results)
        logger.debug(f"coroutine {id} finished")
        return return_val

    @property
    def aggregate(self) -> Call:
        if state_override_supported(self.w3):
            return Call(
                self.multicall_address,
                self.multicall_sig,
                returns=None,
                _w3=self.w3,
                block_id=self.block_id,
                gas_limit=self.gas_limit,
                state_override_code=MULTICALL3_BYTECODE
            )
        
        # If state override is not supported, we simply skip it.
        # This will mean you're unable to access full historical data on chains without state override support.
        return Call(
            self.multicall_address,
            self.multicall_sig,
            returns=None,
            _w3=self.w3,
            block_id=self.block_id,
            gas_limit=self.gas_limit,
        )


class NotSoBrightBatcher:
    """
    This class helps with processing a large volume of large multicalls.
    It's not so bright, but should quickly bring the batch size down to something reasonable for your node.
    """
    def __init__(self) -> None:
        self.step = 10000
    
    def batch_calls(self, calls: List[Call], step: int) -> List[List[Call]]:
        '''
        Batch calls into chunks of size `self.step`.
        '''
        batches = []
        start = 0
        done = len(calls) - 1
        while True:
            end = start + step
            batches.append(calls[start:end])
            if end >= done:
                return batches
            start = end
        
    def split_calls(self, calls: List[Call], unused: None = None) -> Tuple[List[Call],List[Call]]:
        """
        Split calls into 2 batches in case request is too large.
        We do this to help us find optimal `self.step` value.
        """
        center = len(calls) // 2
        chunk_1 = calls[:center]
        chunk_2 = calls[center:]
        return chunk_1, chunk_2
    
    async def rebatch(self, calls):
        # If a separate coroutine changed `step` after calls were last batched, we will use the new `step` for rebatching.
        if self.step <= len(calls) // 2:
            return await run_in_subprocess(self.batch_calls, calls, self.step)
        
        # Otherwise we will split calls in half.
        if self.step >= len(calls):
            new_step = round(len(calls) * 0.99) if len(calls) >= 100 else len(calls) - 1
            logger.warning(f'Multicall batch size reduced from {self.step} to {new_step}. The failed batch had {len(calls)} calls.')
            self.step = new_step
        return await run_in_subprocess(self.split_calls, calls, self.step)


batcher = NotSoBrightBatcher()


def _raise_or_proceed(e: Exception, ct_calls: int, ConnErr_retries: int) -> None:
    """ Depending on the exception, either raises or ignores and allows `batcher` to rebatch. """
    if isinstance(e, aiohttp.ClientOSError):
        if 'broken pipe' not in str(e).lower():
            raise e
        logger.warning(e)
    elif isinstance(e, aiohttp.ClientResponseError):
        strings = ['request entity too large','connection reset by peer']
        if not any([string in str(e).lower() for string in strings]):
            raise e
        logger.warning(e)
    elif isinstance(e, requests.ConnectionError):
        if "('Connection aborted.', ConnectionResetError(104, 'Connection reset by peer'))" not in str(e) or ConnErr_retries > 5:
            raise e
    elif isinstance(e, requests.HTTPError):
        strings = 'request entity too large','payload too large','time-out','520 server error'
        if not any([string in str(e).lower() for string in strings]):
            raise e
        logger.warning(e)
    elif isinstance(e, asyncio.TimeoutError):
        pass
    elif isinstance(e, ValueError):
        if 'out of gas' not in str(e).lower():
            raise e
        if ct_calls == 1:
            raise e
        logger.warning(e)
    else:
        raise e



from functools import lru_cache
from typing import Any, List, Optional, Tuple

# For eth_abi versions < 2.2.0, `decode` and `encode` have not yet been added.
# As we require web3 ^5.27, we require eth_abi compatability with eth_abi v2.0.0b6 and greater.
try:
    from eth_abi import decode, encode
except ImportError: 
    from eth_abi import encode_abi as encode, decode_abi as decode

from eth_typing.abi import Decodable, TypeStr
from eth_utils import function_signature_to_4byte_selector


get_4byte_selector = lru_cache(maxsize=None)(function_signature_to_4byte_selector)

def parse_signature(signature: str) -> Tuple[str, List[TypeStr], List[TypeStr]]:
    """
    Breaks 'func(address)(uint256)' into ['func', ['address'], ['uint256']]
    """
    parts: List[str] = []
    stack: List[str] = []
    start: int = 0
    for end, character in enumerate(signature):
        if character == '(':
            stack.append(character)
            if not parts:
                parts.append(signature[start:end])
                start = end
        if character == ')':
            stack.pop()
            if not stack:  # we are only interested in outermost groups
                parts.append(signature[start:end + 1])
                start = end + 1
    function = ''.join(parts[:2])
    input_types = parse_typestring(parts[1])
    output_types = parse_typestring(parts[2])
    return function, input_types, output_types

def parse_typestring(typestring: str) -> Optional[List[TypeStr]]:
    if typestring == "()":
        return []
    parts = []
    part = ''
    inside_tuples = 0
    for character in typestring[1:-1]:
        if character == "(":
            inside_tuples += 1
        elif character == ")":
            inside_tuples -= 1
        elif character == ',' and inside_tuples == 0:
            parts.append(part)
            part = ''
            continue
        part += character
    parts.append(part)
    return parts


class Signature:
    def __init__(self, signature: str) -> None:
        self.signature = signature
        self.function, self.input_types, self.output_types = parse_signature(signature)
    
    @property
    def fourbyte(self) -> bytes:
        return get_4byte_selector(self.function)

    def encode_data(self, args: Optional[Any] = None) -> bytes:
        return self.fourbyte + encode(self.input_types, args) if args else self.fourbyte

    def decode_data(self, output: Decodable) -> Any:
        return decode(self.output_types, output)



from typing import Any, Callable, Iterable, List, Optional, Tuple, Union

import eth_retry
from eth_typing import Address, ChecksumAddress, HexAddress
from eth_typing.abi import Decodable
from eth_utils import to_checksum_address
from web3 import Web3

from multicall import Signature
from multicall.constants import Network, w3, ASYNC_SEMAPHORE
from multicall.exceptions import StateOverrideNotSupported
from multicall.loggers import setup_logger
from multicall.utils import (chain_id, get_async_w3, run_in_subprocess,
                             state_override_supported)

logger = setup_logger(__name__)

AnyAddress = Union[str,Address,ChecksumAddress,HexAddress]

class Call:
    def __init__(
        self, 
        target: AnyAddress, 
        function: Union[str,Iterable[Union[str,Any]]], # 'funcName(dtype)(dtype)' or ['funcName(dtype)(dtype)', input0, input1, ...]
        returns: Optional[Iterable[Tuple[str,Callable]]] = None, 
        block_id: Optional[int] = None, 
        gas_limit: Optional[int] = None,
        state_override_code: Optional[str] = None, 
        # This needs to be None in order to use process_pool_executor
        _w3: Web3 = None
    ) -> None:
        self.target = to_checksum_address(target)
        self.returns = returns
        self.block_id = block_id
        self.gas_limit = gas_limit
        self.state_override_code = state_override_code
        self.w3 = _w3

        self.args: Optional[List[Any]]
        if isinstance(function, list):
            self.function, *self.args = function
        else:
            self.function = function
            self.args = None

        self.signature = Signature(self.function)
    
    def __repr__(self) -> str:
        return f'<Call {self.function} on {self.target[:8]}>'

    @property
    def data(self) -> bytes:
        return self.signature.encode_data(self.args)

    def decode_output(
        output: Decodable,
        signature: Signature,
        returns: Optional[Iterable[Tuple[str,Callable]]] = None,
        success: Optional[bool] = None
    ) -> Any:
    
        if success is None:
            apply_handler = lambda handler, value: handler(value)
        else:
            apply_handler = lambda handler, value: handler(success, value)

        if success is None or success:
            try:
                decoded = signature.decode_data(output)
            except:
                success, decoded = False, [None] * (1 if not returns else len(returns)) # type: ignore
        else:
            decoded = [None] * (1 if not returns else len(returns)) # type: ignore

        logger.debug(f'returns: {returns}')
        logger.debug(f'decoded: {decoded}')

        if returns:
            return {
                name: apply_handler(handler, value) if handler else value
                for (name, handler), value
                in zip(returns, decoded)
            }
        else:
            return decoded if len(decoded) > 1 else decoded[0]

    @eth_retry.auto_retry
    def __call__(self, args: Optional[Any] = None, _w3: Optional[Web3] = None) -> Any:
        _w3 = self.w3 or _w3 or w3
        args = prep_args(
            self.target,
            self.signature,
            args or self.args,
            self.block_id,
            self.gas_limit,
            self.state_override_code,
        )
        return Call.decode_output(
            _w3.eth.call(*args),
            self.signature,
            self.returns,
        )
    
    def __await__(self) -> Any:
        return self.coroutine().__await__()

    @eth_retry.auto_retry
    async def coroutine(self, args: Optional[Any] = None, _w3: Optional[Web3] = None) -> Any:
        _w3 = self.w3 or _w3 or w3

        if self.state_override_code and not state_override_supported(_w3):
            raise StateOverrideNotSupported(f'State override is not supported on {Network(chain_id(_w3)).__repr__()[1:-1]}.')
        
        async with ASYNC_SEMAPHORE:
            output = await get_async_w3(_w3).eth.call(
                *await run_in_subprocess(
                    prep_args,
                    self.target,
                    self.signature,
                    args or self.args,
                    self.block_id,
                    self.gas_limit,
                    self.state_override_code,
                )
            )

        return await run_in_subprocess(Call.decode_output, output, self.signature, self.returns)
    
def prep_args(
    target: str, 
    signature: Signature, 
    args: Optional[Any], 
    block_id: Optional[int], 
    gas_limit: int, 
    state_override_code: str,
) -> List:

    calldata = signature.encode_data(args)

    args = [{'to': target, 'data': calldata}, block_id]

    if gas_limit:
        args[0]['gas'] = gas_limit

    if state_override_code:            
        args.append({target: {'code': state_override_code}})

    return args



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['multicall_zksync']

package_data = \
{'': ['*']}

install_requires = \
['eth_retry>=0.1.8,<0.2.0', 'web3>=5.28,<6.0']

setup_kwargs = {
    'name': 'multicall-zksync',
    'version': '0.1.1',
    'description': '',
    'long_description': None,
    'author': 'JIVIN0902',
    'author_email': 'jivinvaidya@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)
